% Last and Kubik (1983) algorithm - simple code
clear all; close all; clc;
gamma=6.67E-11; % Newton gravity constant

% opening the file with profile data (2 columns)
[fname,pname] = uigetfile('*.dat','input of data (profile data - ASCII)');
inpXF = dlmread(fname,' ');
% columns are separated into vectors X and F
Xinp=inpXF(:,1);  Finp=inpXF(:,2);
Ninp = length(Xinp); colsinp = length(Xinp); dxinp = Xinp(2)-Xinp(1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUT PARAMETERS - TO BE SET BY THE USER
% number of model layers
rowsinp = 10; %10 for modelled data, 6 for Trnava
% vertical dimension of the model cell
h = 2.22; %2.22 for modelled data, 1.5 for Trnava
% begin of the model space
xmin = 0; ymin = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% various help variables
Minp = rowsinp.*Ninp; M = Minp; d = dxinp; dy = h; dyinp = h; dx = dxinp;

% figure with input profile data
figure, plot(Xinp,Finp);

% generation of the x- and z-coordinates of the calculation points on the model surface
xi = Xinp';
for i = 1:colsinp
    zi(i) = 0;
end;
% generation of the x- and z-coordinates of the line points on the model surface
jj=1;
for j = 1:rowsinp
    for i = 1:colsinp
    xj(jj)= xmin+dx/2+(i-1)*dx;
    zj(jj)= ymin+dy/2+(j-1)*dy; jj = jj + 1; 
    end;
end;
 
% generation of the a-matrix (using effect of 2D rectangular prism)
 for i=1:Ninp
  for j=1:Minp
  r1 = sqrt( (zi(i)+zj(j)-h/2)*(zi(i)+zj(j)-h/2) + (xi(i)-xj(j)+d/2)*(xi(i)-xj(j)+d/2) ); 
  r2 = sqrt( (zi(i)+zj(j)+h/2)*(zi(i)+zj(j)+h/2) + (xi(i)-xj(j)+d/2)*(xi(i)-xj(j)+d/2) );
  r3 = sqrt( (zi(i)+zj(j)-h/2)*(zi(i)+zj(j)-h/2) + (xi(i)-xj(j)-d/2)*(xi(i)-xj(j)-d/2) );
  r4 = sqrt( (zi(i)+zj(j)+h/2)*(zi(i)+zj(j)+h/2) + (xi(i)-xj(j)-d/2)*(xi(i)-xj(j)-d/2) );
  theta1 = atan( (xi(i)-xj(j)+d/2)/(zi(i)+zj(j)-h/2) );
  theta2 = atan( (xi(i)-xj(j)+d/2)/(zi(i)+zj(j)+h/2) );
  theta3 = atan( (xi(i)-xj(j)-d/2)/(zi(i)+zj(j)-h/2) );
  theta4 = atan( (xi(i)-xj(j)-d/2)/(zi(i)+zj(j)+h/2) );
  a(i,j) = 2*gamma*( (xi(i)-xj(j)+d/2)*log((r2*r3)/(r1*r4))+d*log(r4/r3)-...
                     (zj(j)+h/2)*(theta4-theta2)+(zj(j)-h/2)*(theta3-theta1) );
  end; %(for j=1:Minp)
 end; %(for i=1:Ninp) 

 % right-hand matrix of the equation system (interpreted field)
 g = Finp; 

% **************** INVERSION *******************
% simpliest LSQ-solution (not working well, page 716 (Fig. 2b)
lsqdens = a'*inv(a*a')*g;
 jj=1;
   for j = 1:rowsinp
      for i = 1:colsinp
      visu0(j,i)=lsqdens(jj); jj = jj + 1;
      end
   end;
 figure, contourf(flipud(visu0));legend('simplest LSQ inversion');

% simple Tikhonov regularization, little bit better, but still very wrong (shallow)
lam = 0.1; %regularization parameter, usually a positive number between 1E-10 and 1E+10
[mw,nw]=size(a*a'); Id = eye(mw,nw);% genering the identity matrix Id
tichregdens = a'*inv(a*a'+lam*Id)*g;
 jj=1;
   for j = 1:rowsinp
      for i = 1:colsinp
      visu0(j,i)=tichregdens(jj); jj = jj + 1;
      end
   end;
 figure, contourf(flipud(visu0));legend('Tikhonov inversion without compacting');
 
%Last-Kubik 1983 algorithm (simpliest version)
% preparation:
l_0 = 0.01; % THIS CAN BE CHANGED BY USER, estimated noise/signal ratio 
eps = 1E-3; % THIS CAN BE CHANGED BY USER, value 1E-11, recommended in page 715, right column, central part
w_ny = eye(M,M);   w_ny_inv = inv(w_ny); % genering the identity matrix Id as the first iteration for the weighting matrix w_ny
w_e_inv = l_0*l_0*a*w_ny_inv*a'; % weighting (noise e) matrix in Eq.11 an Eq.17

% number of iterations
number_iterations = 10;
for k=1:number_iterations % iteration cycle
   disp('interation nr.:'); k    
   LKdens = w_ny_inv*a'*inv(a*w_ny_inv*a'+w_e_inv)*g; %main solution, Eq. 15
   LKdensT = LKdens';
   % results visualisation after each iteration step
   jj=1;
   for j = 1:rowsinp
      for i = 1:colsinp
      visu1(j,i)=LKdensT(jj); jj = jj + 1;
      end
   end;
   infoonmethod='Last and Kubik, it.Nr.: '; fordisplay=append(infoonmethod, num2str(k));
   figure, contourf(flipud(visu1)); legend(fordisplay);
   % adapting the weighting matrix w_ny, Eq.16
   for j=1:M
       w_ny_inv(j,j) = LKdens(j)*LKdens(j)+eps; %empiricky overene, ze takto vyratana inverzna je identicka s tou vyratanou podla rovnice (8) a potom invertovanou
   end; %(for j=1:M)
      w_e_inv = l_0*l_0*a*w_ny_inv*a'; % adapting the weighting matrix w_ne, Eq.11
end; %(for k=1:10)

% exporting results into an ASCII SRF grid
xmax = xj(M); ymax = zj(M); %help variables for solution grid output
grid_write_SRF_ASCII(visu1,xmin,xmax,ymin,ymax,'LKdens.grd');

